/* *******************************************************
  Anthony Garratt, Donald Robertson and Stepehn Wright"Permanent vs. 
  Transitory Components and Economic Fundamentals" forthcoming Journal of 
  Applied Econometrics 
**************************************************************************** */
new;
closeall;
format 8,4;
library pgraph;graphset;
output file = bnsv.out reset;
time;


freq_op = 2;
porder = 2;
case = 4;
xcase=1;

/* ************************************
*   STEP 1-1: Data loading and handling
*  ************************************ */
@* Data loading @


/*     ******  Data loading ********* 
*  data in the column order of y, ys, r, rs, e, hy, po, dpo, dp, ppsp 
*  data 1963q1-1999q4, 148 observations
*  estimation period 65q1-99q4, 140 observations 
*  number of initial values discarded= 8 ****     */

load ukmod99[148,10]=c:\MyDoc\bncopap\gauss\ukmod99.dat;



@ number of initial values discarded @
stt = 8;  @ 65q1-99q1 @ 


@* Endogenous I(1) variables @

ppsp = ukmod99[.,10];  @ log of domestic price - log of foreign price @
exch = ukmod99[.,5];  @ log of effective exchange rate @
r    = ukmod99[.,3];  @ log of (1+domestic interest) @
rs   = ukmod99[.,4];  @ log of (1+foreign interest) @
yd   = ukmod99[.,1];  @ log of domestic output @
ys   = ukmod99[.,2];  @ log of foreign output @
hy   = ukmod99[.,6];  @ log of (M0/output) @
dp   = ukmod99[.,9];  @ inflation @


@* Exogenous I(1) variables @

po  = ukmod99[.,7];   @ log of oil price @
dpodata = ukmod99[.,8];   @ oil price inflation @

@* Selection of I(1) endogenous variables, Y. @

Ydata = rs~exch~r~dp~yd~ppsp~hy~ys; @ same order as in EJ paper @

@* Selection of I(1) exogenous variables X. @

Xdata = po; @ 1 exo @

yorder = 8;  @ number of endo I(1) variables @
xorder = 1;  @ number of exo I(1) variables @
Worder = 0;  @ number of exo I(0) variables @

/* *******************
*  STEP 2: Estimation
*  ******************* */

@*  Long-run cointegrating vectors
*   Over-identifying restrictions for Theory Model
*
*   Nine I(1) variables with restricted trend and 5 cointegrations,
*   so the dimension of beta(star) is 10 by 5.
*
*   Full (34) restrictions with 18 over-identifying restrictions
*   to be tested is as follows:
*
*            PPP   IRP  Gap  Real FIS
*   beta = {   0    0    0    0    0   PO
*             -1    0    0    0    0   E
*              0   -1    0    0    0   RS
*              0    1    0    *    1   R
*              0    0    0    0   -1   DP
*              0    0    1    0    0   Y
*              1    0    0    0    0   PPSP
*              0    0    0    1    0   HY
*              0    0   -1    0    0   YS
*              0    0    0    *    0}  TR
*
*   Also consider the exactly identifying restrictions
*******************************************  *@
output on; time;

@ ESTIMATED BETA  FIVE CV OVER IDENTIFYING RESTRICTIONS 65Q1-99Q4 @

ov5beta = zeros(10,5);  
@ PPP: pps - e @
   ov5beta[7,1] = 1;    ov5beta[3,1] = -1;
@ UIP: r - rs @
   ov5beta[4,2] = 1;    ov5beta[2,2] = -1;
@ Gap: y - ys @
   ov5beta[6,3] = 1;    ov5beta[9,3] = -1;
@ Real balance: hy + a1*r + a2*t @
   ov5beta[8,4] = 1;
   ov5beta[4,4] = 56.0975;
   ov5beta[10,4] = 0.007319;
@ Fisher relation: r - dp @
   ov5beta[4,5] = 1;    ov5beta[5,5] = -1;




   Y = Ydata; X = Xdata;
   Z = Xdata~Ydata;   ACTZ = Z;

@* Initial values at T, T-1, T-2 for forecast @

   ZT  = Z[rows(Z),.];   ZT = ZT';   @ 9 by 1 @
   ZT1 = Z[rows(Z)-1,.]; ZT1 = ZT1'; @ Z_t-1 @
   ZT2 = Z[rows(Z)-2,.]; ZT2 = ZT2'; @ Z_t-2 @
   ZT3 = Z[rows(Z)-3,.]; ZT3 = ZT3'; @ Z_t-3 @
   ZT4 = Z[rows(Z)-4,.]; ZT4 = ZT4'; @ Z_t-4 @

   DZ   = Z - lag(Z);       ACTDZ = DZ;
   DZT  = DZ[rows(DZ),.];   DZT = DZT';   @ 9 by 1 @

@* Initial values at 0 for bootstrap @

   Z0  = Z[stt,.];   Z0  = Z0';    @ 9 by 1 @
   DZ0 = DZ[stt,.];  DZ0 = DZ0';   @ 9 by 1 @
   ZINI = Z[1:stt,.]; ZINI = ZINI';   @ 9 by stt @

   dpo = dpodata[stt+1:rows(dpodata)];
   dpo0 = dpodata[stt];

@ ********************
* Data transformation:
* ********************
* dimensions of the resuling matrix:
* DY: my by T,  DX: mx by T,  Z1: m by T, ADZLAGS: m(maxp-1) by T
* **************************************************************** @

   { DY, Z1, DX, ADZLAGS } = TRDATA(Y,X);

   m = xorder + yorder;
   p = porder;
   DY = DY[.,stt+1:cols(DY)];
   sample = cols(dy);
   Z1 = Z1[.,stt+1:cols(Z1)];

   DX = DX[.,stt+1:cols(DX)];
   ADZLAGS = ADZLAGS[.,stt+1:cols(ADZLAGS)]; @ up to maximum p @

   if p eq 1;
   else; DZLAGS = ADZLAGS[1:m*(p-1),.];
   endif;

   if worder eq 0;  W = 0; WWP = 0;
   else; W = W';   @ Transformation conformable to DY @
         W = W[.,stt:cols(W)];
   endif;

/* **********************************************
*  STEP 2: The long-run cointegration estimation
*  **********************************************
*  We assume that you have already estimated ML estimates of
*  the cointegrating vectors which are obtained via e.g. Micorfit. */

@* construction of zstar(-1):  @

   const = ones(sample,1);     const = const';
   
   t=cols(z1);
   trend1 = seqa(-1,1,t);  trend1 = trend1';

   ZS1 = Z1;
   IF     case eq 2;  ZS1 = Z1|CONST;
   Elseif case eq 4;  ZS1 = Z1|TREND1;  @ 9+1 by T @
   ENDIF;

@* Estimating the cointegrating vectors @
@* only for exact-identified case @
@* over-identified beta is fixed @

@* Cointegration test and Johansen eigenvectors *@
   W=0;
   { COTEST, Jbeta } =
   COINTEST(DY,DX,ZS1,DZLAGS,W,CASE,yorder,xorder,worder,porder);

/*
"";"COINTEGRATION TEST RESULTS:";"";"eigenval   trace     max";
format /rd 8,4; COTEST;"";"press any key"; wait;
*/


@* EC terms: cointegrating relations: r by T @



   ov5xi = ov5beta'ZS1; @ 5 by T @






/*  ************************************
*   STEP 3: Short-run Dynamic Estimation
*   ************************************ */
/*  *********************************************
*   STEP 3(a). Conditional VEC estimation for dy
*   *********************************************  */



@* Over-identified case @
@* r = 5 @
   YY = ov5xi|DZLAGS|DX|const;   @ 5+9+1+1 by T @
   kpara = yorder*rows(YY) + (10*5-5*5) -23;
   { ov5ally, ov5ey, ov5omegy, ov5LL, ov5AIC, ov5SBC, ov5HQ }
   = VECdy(YY,DY,sample,yorder,kpara);







save ov5ally ov5ey ov5omegy ov5LL ov5AIC ov5SBC ov5HQ ;



/*  ******************************************
*   STEP 3(b). Marginal VAR estimation for dx
*   ******************************************  */
@* Estimation results for dpo: 65q1-...
*  Two options:
*   case (a). dpo on intercept
*   case (b). dpo on intercept, dz(-1) @

@if xcase eq 1;@

@* case (a) @
   regx = const; regx = regx';  @ T by 1 @
   { allxa, exa, omegaxa, LLxa, AICxa, SBCxa, HQxa }
   = MARdx(regx,dpo,sample);

@   else;@

@* case (b) @
   regx = const|dzlags; regx = regx';  @ T by 1+9 @
   { allxb, exb, omegaxb, LLxb, AICxb, SBCxb, HQxb }
   = MARdx(regx,dpo,sample);

@endif;@


/*  ******************************************
*   DEFINING CORRECT TREND SPECIFICATION WHEN
    DEFINING CVDAT IN BNSV PROCEDURE 
*   ******************************************  */


trdov5={0, 0, 0, -0.007319,0};





/*  ************************************************************************
*   CALCULATING BN TRENDS AND CYCLES 
     FOR THE FULL RANGE OF MODELS 
*   *************************************************************************  */



@* Over-identified case, r=5  @

{ov5tr,ov5cy,ov5trecm,ov5trdel,ov5cydel,ov5cyecm,
ov5cecm,ov5forl,ov5y, ov5dy,ov5z,ov5totu,ov5paru,ov5futu,ov5gr,ov5cc,bn}
 = BNSV(ov5ally,allxa,allxb,ov5omegy,omegaxa,omegaxb,
    yorder,xorder,ydata,xdata,ov5beta,stt,porder,trdov5);

save ov5tr ov5cy ov5trecm ov5trdel ov5cydel ov5cyecm;
save ov5cecm ov5forl ov5y  ov5dy ov5z ov5totu ov5paru ov5futu ov5gr ov5cc;





END;




/* ***********
   Procedures
   *********** */

/* ***********************  */
   PROC (4) = TRDATA(Y,X);
/* ***********************
*   Generate the appropriate Data matrices of DY, DX, Z1, DZLAGS,
*   to be used in the VEC model, depending on the order of VAR(p)
*
*   DY: First difference of endogenous I(1) variables
*   DX: First difference of exogenous I(1) variables
*   Z = X~Y
*   Z1: First lagged values of Z
*   ADZLAGS = DZ1~DZ2~...~DZ(p-1)
*   ******************************************************************* */
Local  MM,Y1,DY,Z,Z1,adzlags,dX,X1,
       DZ,DZ1,DZ2,DZ3,DZ4,DZ5,DZ6,DZ7,DZ8,DZ9,DZ10,DZ11;

Y1 = LAG(Y); DY = Y - Y1;

IF Xorder eq 0;   Z = Y;        @ No exog variables: T by m @
ELSE;             Z = X~Y;      @ T by m @
                  X1 = LAG(X);  DX = X - X1;
ENDIF;

Z1 = LAG(Z);       DZ = Z - Z1;      DZ1 = Lag(DZ);    DZ2 = Lag(DZ1);
DZ3 = Lag(DZ2);    DZ4 = Lag(DZ3);   DZ5 = Lag(DZ4);   DZ6 = Lag(DZ5);
DZ7 = Lag(DZ6);    DZ8 = Lag(DZ7);   DZ9 = Lag(DZ8);   DZ10 = Lag(DZ9);
DZ11 = Lag(DZ10);

@ all lagged values up to maximum lag order @

ADZLAGS = DZ1~DZ2~DZ3~DZ4~DZ5~DZ6~DZ7~DZ8~DZ9~DZ10~DZ11;

@* transpose the data for matching the specification *@

DY = DY';  Z1 = Z1';
DX = DX';  ADZLAGS = ADZLAGS';

RETP(DY,Z1,DX,ADZLAGS);ENDP;

/* ****************************************************** */
   PROC (2) = COINTEST(DY,DX,Z1,DZLAGS,W,CASE,my,mx,mw,p);
/* ******************************************************
*   (1)  OLS of DY on YY to obtain R0 and then calculate
*
*       S00 = (1/T)R0*R0'
*
*   (2)  OLS of Z1 on YY to obtain R1 and then calculate
*
*       S11 = (1/T)R1*R1',      S01 = (1/T)R0*R1'
*
*   (3) Then, calculate
*
*       Bmatirx = S11
*       Amatrix = S11 - S01'Invpd(S00)*S01
*
*    For further details see Johansen (1988, 1991) and PSS2 (1997)
*  ***************************************************************** */
local  sample,URPAI,inpt,trend,dreg,yy,invYY,invS00,Amatrix,Bmatrix,
       S00,S01,S11,C0,C1,R0,R1,zorder,testres,pp,va,ve,dd,zi,sumva,
       maxva,Ltrace,Lmax,sumv,cvmax,cvtrace,vva,vaa,im1,im2,jbeta;

sample = cols(dy);
inpt = ones(sample,1);     inpt = inpt';
trend = seqa(1,1,sample);  trend = trend';

@* Z1 = (X1 Y1) and any deterministic regressors @

zorder = rows(z1);  @ dimension of z1 @

@  Construction of all other stationary regressors, YY @

if p eq 1; YY = 0;
else;      YY = DZLAGS;
endif;

if YY eq 0;
   if mx eq 0;
   else;  YY = DX;
   endif;
else;
   if mx eq 0;
   else;  YY = YY|DX;
   endif;
endif;

if YY eq 0;
   if mw eq 0;
   else; YY = W;
   endif;
else;
   if mw eq 0;
   else; YY = YY|W;
   endif;
endif;

if YY eq 0;
   IF     case eq 3;  YY = inpt;
   ELSEIF case eq 4;  YY = inpt;
   ELSEIF case eq 5;  YY = inpt|trend;
   ENDIF;
else;
   IF     case eq 3;  YY = YY|inpt;
   ELSEIF case eq 4;  YY = YY|inpt;
   ELSEIF case eq 5;  YY = YY|inpt|trend;
   ENDIF;
endif;

@ Computation of S00, S11 and S10 @

IF YY eq 0; R0 = DY; R1 = Z1;
ELSE;
     InvYY = Invpd(YY*YY');
     C0 = DY*YY'InvYY;
     R0 = DY - C0*YY;
     C1 = Z1*YY'InvYY;
     R1 = Z1 - C1*YY;
ENDIF;

S00 = R0*R0'./sample;
S11 = R1*R1'./sample;
S01 = R0*R1'./sample;

@ Computation of Amatrix and Bmatrix @

InvS00 = invpd(S00);
Amatrix = S11 - S01'InvS00*S01;
Bmatrix = S11;

@ Unrestricted pi matrix @

URPAI = S01*invpd(S11);

/* *******************************************************
   Calculation of Johansen Beta's using eigenvalue routine
   ******************************************************* */
pp = chol(S11);
dd = inv(pp')*S01'invS00*S01*inv(pp);

{va,ve} = EIGHV(dd);

Jbeta = zeros(zorder,zorder);
zi = 1;
do until zi > zorder;

   Jbeta[.,zi] = inv(pp)*ve[.,zi];
   zi = zi + 1;

endo;

vva = zeros(zorder,1);
zi = 1;
do until zi > zorder;

   vva[zi,1] = va[zorder-zi+1,1];
   zi = zi + 1;

endo;

/* **********************************************************
   Computation of the Trace and the Max Eigenvalue statistics
   ********************************************************** */
vaa = zeros(zorder,1);
sumv = 0;
zi = 1;
do until zi > zorder;

   vaa[zi,1] = ln(1 - va[zi,1]);
   sumv = sumv + vaa[zi,1];
   zi = zi + 1;
endo;

sumva = zeros(my,1);
maxva = zeros(my,1);
sumva[1,1] = sumv;
maxva[1,1] = ln(1 - va[zorder,1]);
zi = 2;
do until zi > my;

   im1 = zorder + 2 - zi;
   sumva[zi,1] = sumva[zi-1,1] - vaa[im1,1];
   im2 = zorder + 1 - zi;
   maxva[zi,1] = ln(1 - va[im2,1]);

   zi = zi + 1;
endo;

Ltrace = zeros(my,1);  Lmax = zeros(my,1);
zi = 1;
do until zi > my;

   Ltrace[zi,1] = -sample*sumva[zi,1];
   Lmax[zi,1] = -sample*maxva[zi,1];
   zi = zi + 1;
endo;

VVA = VVA[1:my];
TESTRES = VVA~LTRACE~LMAX;

RETP(TESTRES,Jbeta);
ENDP;

/* ************************************ */
   Proc (7) = VECdy(YY,DY,sample,my,k);
/* ************************************ */
Local ally,ey,omegay,AIC,SBC,HQ,LL;

ally = DY*YY'*INV(YY*YY'); @ my by r+9+1+1 @
ey = DY - ally*YY;         @ residuals for dy: yorder by T @
omegay = ey*ey'./(sample-rows(YY));
omegay = ey*ey'./(sample-rows(YY)-1); @ Mfit: restricted trend? @

LL = -(sample/2)*ln(det(ey*ey'./sample))-(sample*my/2)*(1+ln(2*pi));

/* k = rows(ally)*cols(ally); */
AIC = LL - k;
SBC = LL - (k/2)*ln(sample);
HQ  = LL - k*ln(ln(sample));

RETP(ally,ey,omegay,LL,AIC,SBC,HQ);ENDP;

/* ******************************** */
   Proc (7) = MARdx(regx,dx,sample);
/* ******************************** */
Local allx,ex,omegax,AIC,SBC,HQ,LL,k;

allx = inv(regx'regx)*(regx'dx);
ex   = dx - regx*allx;            @ T by 1 @
omegax = ex'ex/(sample-cols(regx));

LL = -(sample/2)*ln(ex'ex/sample)-(sample/2)*(1+ln(2*pi));
k = rows(allx)*cols(allx);
AIC = LL - k;
SBC = LL - (k/2)*ln(sample);
HQ  = LL - k*ln(ln(sample));

RETP(allx,ex,omegax,LL,AIC,SBC,HQ);ENDP;




/* ************************************************** */
   Proc (14) = weights(EX0A,EX0B,EX1A,EX1B,EX2A,EX2B,
             EX3A,EX3B,EX4A,EX4B,EX5A,EX5B,OV5A,OV5B);
/* ************************************************** */
Local EX0AW,EX0BW,EX1AW,EX1BW,EX2AW,EX2BW,EX3AW,EX3BW,EX4AW,EX4BW,
      EX5AW,EX5BW,OV5AW,OV5BW,wtmp,maxtmp,etmp,sum;

wtmp = zeros(14,1);
wtmp[1]  = EX0A; wtmp[2]  = EX0B;
wtmp[3]  = EX1A; wtmp[4]  = EX1B;
wtmp[5]  = EX2A; wtmp[6]  = EX2B;
wtmp[7]  = EX3A; wtmp[8]  = EX3B;
wtmp[9]  = EX4A; wtmp[10] = EX4B;
wtmp[11] = EX5A; wtmp[12] = EX5B;
wtmp[13] = OV5A; wtmp[14] = OV5B;

maxtmp = maxc(wtmp).*ones(14,1);

@* to avoid exp computational overflow, subtract the maximum @

etmp = exp(wtmp-maxtmp);

sum = sumc(etmp);

EX0AW = etmp[1]/sum;  EX0BW = etmp[2]/sum;
EX1AW = etmp[3]/sum;  EX1BW = etmp[4]/sum;
EX2AW = etmp[5]/sum;  EX2BW = etmp[6]/sum;
EX3AW = etmp[7]/sum;  EX3BW = etmp[8]/sum;
EX4AW = etmp[9]/sum;  EX4BW = etmp[10]/sum;
EX5AW = etmp[11]/sum; EX5BW = etmp[12]/sum;
OV5AW = etmp[13]/sum; OV5BW = etmp[14]/sum;

RETP(EX0AW,EX0BW,EX1AW,EX1BW,EX2AW,EX2BW,EX3AW,EX3BW,EX4AW,EX4BW,
EX5AW,EX5BW,OV5AW,OV5BW);ENDP;






/* ********************************************************** */
   Proc (4) = SCOVA(omegay,omegaxa,omegaxb,my,mx,ey,exa,exb);
/* ********************************************************** */
local AOMEGA,BOMEGA,AEZ,BEZ,o1,o2;

o1 = omegaxa~zeros(mx,my);
o2 = zeros(my,mx)~omegay;
aomega = o1|o2;

o1 = omegaxb~zeros(mx,my);
bomega = o1|o2;

aez = exa'|ey;  @ residuals: 9 by T @
bez = exb'|ey;

RETP(AOMEGA,BOMEGA,AEZ,BEZ);ENDP;


/* ************************************************************* */
Proc(17)= BNSV(cally,callxa,callxb,omegay,omegaxa,
              omegaxb,yorder,xorder,ydata, xdata,beta, 
              stt,porder,trends);           
/* ************************************************************* */

local y,t,n,r,h,dswitch,deltasize,varlist,
      alpha,phi,deltazero,b0,psiy0,psi,o1,o2,
      aomega,sigma,g,delta,pim,bigpsi,a1,a2,cvdat,
      dy,z,d,jm,bigv,sigmav,e,yss,growth,cconst,
      zt,forcz,forclevels,i,futunc,func,ddd,j,
      parunc,dvecd,bstar,jmat,sigdelta,zzz,fff,punc,
      aaa,dpi1j,dj,forct,forczero,ally,alldy,allt,
      alltt,allz,allcv,meanzless1,zless1,ztwid,bn, 
      ydev,bntrend,bndelta,bnecm,allcvt,ydevdelta,
      ydevecm,totunc;


@ DATA @

y=xdata~ydata; @ 148 by 10 @
@ sample size in estimation @
t=rows(y)-(stt-porder);
@ stt=obs dropped for start date, porder=lags in levels @
y=y[(stt-porder)+1:rows(y),.];

@ GENERAL SETTINGS @ 

@ number of variables in underlying VAR- 
  NB does not include time trend @
n=yorder+xorder;  
@ number of cointegrating vectors @
r=cols(beta);  
@forecast horizon@
h=40; 
@ switch for data-based forecast (dswitch=1), 
  or mean-based forecast (dswitch=0) @
dswitch=1; 
@ implied size of delta assuming single system growth rate @
deltasize=n*(n+r+1);

varlist="po~rs~exch~ret~dp~yd~ppsp~hy~ys"; 
@ecmlist="ppp,r-r*,y-y*,money,r-pi";@



@ COEFFICIENT MATRICES REQUIRED TO COMPUTE FORECASTS @

alpha=zeros(1,r)|cally[.,1:r];
phi=zeros(1,n)|cally[.,(r+1):(n+r)];
deltazero=callxa;
psiy0=cally[.,(n+r+1)];
b0=cally[.,(n+r+2)];

beta=beta[1:n,.];


@ constructing the 9 by 1 vector, psi, of constants
  in equation (12) in the note. @  
psi=deltazero|((psiy0*deltazero)+b0);

@ COVARIANCE MATRIX OF RESIDUALS  
  including oil price, as in proc SCOVA @
o1 = omegaxa~zeros(xorder,yorder);
o2 = zeros(yorder,xorder)~omegay;
aomega = o1|o2;
sigma= aomega;

@ SIZE OF Z SYSTEM @
g=n+r+1;

@ DELTA STACKS ALL ESTIMATED PARAMETERS @

delta=(psi'~vec(phi)'~vec(alpha)')';
pim=alpha*beta';
bigpsi=psi;


@ CALCULATES AR2 REPRESENTATION x(t)=bigpsi+a1*x(t-1)+a2*x(t-2)+e(t)  @

a1=phi+eye(n)+pim;
a2=-phi;

@ CONSTRUCTS DATA FOR Z MATRIX @

cvdat=(y[1:t-1,1:n]*beta)-seqa(-1,1,t-1).*trends';
dy=zeros(t-1,n);
dy[1:t-1,1:n]=(y[2:t,1:n]-y[1:t-1,1:n]);
z=ones(t-1,1)~dy~cvdat;

@ we will need meancv, but need to derive e1 from bigpsi, alpha, etc @
@meancv=(e1.*ones(t-1,1));@

d=zeros(g,g);
d[1:g,1]=(ones(1,1)|(bigpsi-alpha*trends)|(-trends));
d[2:(2+n-1),2:(2+n-1)]=phi+pim;
d[n+2:g,2:(2+n-1)]=beta';
d[2:(2+n-1),n+2:g]=alpha;
d[n+2:g,n+2:g]=eye(r);

jm=zeros(n,g);
jm[1:n,2:2+n-1]=eye(n);


@ STATIONARY SYSTEM IS: zt=Dz(t-1) + vt, where vt=bigv*epsilont @
bigv=zeros(g,n);
bigv[2:2+n-1,1:n]=eye(n);
sigmav=bigv*sigma*bigv';


@ CALCULATE IMPLICIT GROWTH VECTOR AND COINTEGRATING CONSTANT @
@ first calculate E matrix  for system y(t)=psiy+Ey(t-1) +w(t)@
e=d[2:g,2:g];
yss=inv(eye(g-1)-e)*d[2:g,1];
growth=yss[1:n,.];
cconst=yss[n+1:g-1,.];


@  CALCULATE MEAN FORECASTS OF Z AND LEVELS OF VAR VARIABLES @

@ if d=1, pick up last element of data Note this is t-1 as z 
  is a shortened vector due to differencing @

if dswitch;
zt=z[t-1,.];
else;

@ if d=0, forecast from system steady state @
zt=1~yss';
endif;

forcz=zeros(h+1,g);
forclevels=zeros(h+1,n);
forcz[1,.]=zt;

@pick out terminal sample values of var variables as 1st obs in forecast@

forclevels[1,1:n]=y[t,.];
i=2;
do while i<=h+1;
forcz[i,.]=(d*forcz[i-1,.]')';
forclevels[i,.]=forclevels[i-1,.]+(jm*forcz[i,.]')';
i=i+1;
endo;


@ CALCULATES FUTURE UNCERTAINTY AS DOUBLE SUM @
futunc=zeros(h,n);
func=zeros(n,n);
i=0;
do while i<=h-1;
ddd=zeros(g,g);
j=0;
do while j<=i;
ddd=ddd+power(d,j,g);
j=j+1;
endo;
func=jm*ddd*sigmav*ddd'jm'+func;
i=i+1;
futunc[i,1:n]=diag(func)';
endo;


@ CALCULATES PARAMETER UNCERTAINTY  @

parunc=zeros(h,n);
dvecd=zeros(g^2,deltasize);

bstar=eye(g);
bstar[2:2+n-1,n+2:n+r+1]=beta;

jmat=zeros(g,n);
jmat[2:2+n-1,1:n]=eye(n);
dvecd=bstar.*.jmat;

@ CALCULATES VARIANCE-COVARIANCE OF ESTIMATED PARAMETERS @

sigdelta=zeros(n*g,n*g);
zzz=ones(t-2,1)~dy[1:t-2,.]~cvdat[2:t-1,.];
fff=zzz[1:t-2,.].*.eye(n);
sigdelta=inv(fff'(eye(t-2).*.inv(sigma))*fff);

parunc=zeros(h,n);
punc=zeros(n,n);

i=1;
aaa=zeros(n,deltasize);
do while i<=h;
j=0;
ddd=zeros(g*g,g*g);
do while j<=i-1;
dpi1j=power(d',i-1-j,g);
dj=power(d,j,g);

ddd=ddd+dpi1j.*.dj;
j=j+1;
endo;

aaa=aaa+((zt).*.jm)*ddd*dvecd;

punc=aaa*(sigdelta)*aaa';
parunc[i,.]=(diag(punc))';
i=i+1;
endo;

totunc=futunc+parunc;



@ CALCULATES BN/CCE TRENDS @

@ forecast time axis: final period of history, then h periods of forecast@
forct=seqa(1,1,h+1);
forczero=zeros(h+1,1);

@ history + forecast @
ally=(y[1:t-1,.]'~forclevels[1:h+1,1:n]')';
alldy=ally[2:(t+h),.]-ally[1:(t+h)-1,.];
allt=seqa(1965,.25,(t+h));
alltt=seqa(1,1,(t+h)-1);
allz=(z[1:t-2,.]'~forcz')';
allcv=allz[.,g-r+1:g];

meanzless1=yss';

@ y=[delta(x),beta'x(t-1)]'@
zless1=allz[.,2:g];  

@ y=y-mean(y)=y-[growth,coin constants]@
ztwid=zless1-meanzless1;

jm=zeros(n,g-1);
jm[1:n,1:n]=eye(n);
bn=jm*e*inv(eye(g-1)-e); 


@ Beveridge Nelson Trend Deviations @
ydev=-(bn*ztwid')'; 

@ Beveridge Nelson Trends @
bntrend=ally[2:(t+h),.]-ydev; 
@NB bntrend*beta=cconst'@

bndelta=bn[.,1:n]; @ matrix to derive contribution of deltas to BN  Deviations @  
bnecm=bn[.,n+1:n+r]; @ matrix to derive ontribution of lagged ECM terms to BN Deviations @
bndelta=bndelta-bnecm*beta'; @ re-jigged so that ecm terms are current dated@
allcvt=ally[2:(t+h),.]*beta-seqa(0,1,t+h-1).*trends'; @ current, rather than lagged ecm terms @
ydevdelta=(bndelta*(alldy-growth')')'; @ contribution of delta terms to BN deviations@
ydevecm=(bnecm*(allcvt-cconst')')';@ contribution of current-dated ecm to BN deviations@

RETP(bntrend,ydev,bnecm,bndelta,ydevdelta,ydevecm,allcvt,forclevels,
     ally,alldy,allz,totunc,parunc,futunc,growth,cconst,bn);
ENDP;


/* ************************************************************* */
Proc(9)= R0BNSV(cally,callxa,callxb,omegay,omegaxa,
              omegaxb,yorder,xorder,ydata, xdata, 
              stt,porder);           
/* ************************************************************* */

local y,t,n,r,h,dswitch,deltasize,varlist,
      alpha,phi,deltazero,b0,psiy0,psi,o1,o2,
      aomega,sigma,g,delta,pim,bigpsi,a1,a2,cvdat,
      dy,z,d,jm,bigv,sigmav,e,yss,growth,cconst,
      zt,forcz,forclevels,i,futunc,func,ddd,j,
      parunc,dvecd,bstar,jmat,sigdelta,zzz,fff,punc,
      aaa,dpi1j,dj,forct,forczero,ally,alldy,allt,
      alltt,allz,allcv,meanzless1,zless1,ztwid,bn, 
      ydev,bntrend,bndelta,bnecm,allcvt,ydevdelta,
      ydevecm,totunc;


@ DATA @

y=xdata~ydata; @ 148 by 10 @
@ sample size in estimation @
t=rows(y)-(stt-porder);
@ stt=obs dropped for start date, porder=lags in levels @
y=y[(stt-porder)+1:rows(y),.];

@ GENERAL SETTINGS @ 

@ number of variables in underlying VAR- 
  NB does not include time trend @
n=yorder+xorder;  
@forecast horizon@
h=40; 
@ switch for data-based forecast (dswitch=1), 
  or mean-based forecast (dswitch=0) @
dswitch=1; 
@ implied size of delta assuming single system growth rate @
deltasize=n*(n+1);

varlist="po~rs~exch~ret~dp~yd~ppsp~hy~ys"; 
@ecmlist="ppp,r-r*,y-y*,money,r-pi";@



@ COEFFICIENT MATRICES REQUIRED TO COMPUTE FORECASTS @

phi=zeros(1,n)|cally[.,1:n];
deltazero=callxa;
psiy0=cally[.,(n+1)];
b0=cally[.,(n+2)];



@ constructing the 9 by 1 vector, psi, of constants
  in equation (12) in the note. @  
psi=deltazero|((psiy0*deltazero)+b0);

@ COVARIANCE MATRIX OF RESIDUALS  
  including oil price, as in proc SCOVA @
o1 = omegaxa~zeros(xorder,yorder);
o2 = zeros(yorder,xorder)~omegay;
aomega = o1|o2;
sigma= aomega;

@ SIZE OF Z SYSTEM @
g=n+1;

@ DELTA STACKS ALL ESTIMATED PARAMETERS @

delta=(psi'~vec(phi)')';
bigpsi=psi;


@ CALCULATES AR2 REPRESENTATION x(t)=bigpsi+a1*x(t-1)+a2*x(t-2)+e(t)  @

a1=phi+eye(n);
a2=-phi;

@ CONSTRUCTS DATA FOR Z MATRIX @

dy=zeros(t-1,n);
dy[1:t-1,1:n]=(y[2:t,1:n]-y[1:t-1,1:n]);
z=ones(t-1,1)~dy;

@ we will need meancv, but need to derive e1 from bigpsi, alpha, etc @
@meancv=(e1.*ones(t-1,1));@

d=zeros(g,g);
d[1:g,1]=ones(1,1)|bigpsi;
d[2:(2+n-1),2:(2+n-1)]=phi;


jm=zeros(n,g);
jm[1:n,2:2+n-1]=eye(n);


@ STATIONARY SYSTEM IS: zt=Dz(t-1) + vt, where vt=bigv*epsilont @
bigv=zeros(g,n);
bigv[2:2+n-1,1:n]=eye(n);
sigmav=bigv*sigma*bigv';


@ CALCULATE IMPLICIT GROWTH VECTOR AND COINTEGRATING CONSTANT @
@ first calculate E matrix  for system y(t)=psiy+Ey(t-1) +w(t)@
e=d[2:g,2:g];
yss=inv(eye(g-1)-e)*d[2:g,1];
growth=yss[1:n,.];



@  CALCULATE MEAN FORECASTS OF Z AND LEVELS OF VAR VARIABLES @

@ if d=1, pick up last element of data Note this is t-1 as z 
  is a shortened vector due to differencing @

if dswitch;
zt=z[t-1,.];
else;

@ if d=0, forecast from system steady state @
zt=1~yss';
endif;

forcz=zeros(h+1,g);
forclevels=zeros(h+1,n);
forcz[1,.]=zt;

@pick out terminal sample values of var variables as 1st obs in forecast@

forclevels[1,1:n]=y[t,.];
i=2;
do while i<=h+1;
forcz[i,.]=(d*forcz[i-1,.]')';
forclevels[i,.]=forclevels[i-1,.]+(jm*forcz[i,.]')';
i=i+1;
endo;



@ CALCULATES BN/CCE TRENDS @

@ forecast time axis: final period of history, then h periods of forecast@
forct=seqa(1,1,h+1);
forczero=zeros(h+1,1);

@ history + forecast @
ally=(y[1:t-1,.]'~forclevels[1:h+1,1:n]')';
alldy=ally[2:(t+h),.]-ally[1:(t+h)-1,.];
allt=seqa(1965,.25,(t+h));
alltt=seqa(1,1,(t+h)-1);
allz=(z[1:t-2,.]'~forcz')';


meanzless1=yss';

@ y=[delta(x),beta'x(t-1)]'@
zless1=allz[.,2:g];  

@ y=y-mean(y)=y-[growth,coin constants]@
ztwid=zless1-meanzless1;

jm=zeros(n,g-1);
jm[1:n,1:n]=eye(n);
bn=-jm*e*inv(eye(g-1)-e); 


@ Beveridge Nelson Trend Deviations @
ydev=(bn*ztwid')'; 

@ Beveridge Nelson Trends @
bntrend=ally[2:(t+h),.]-ydev; 

bndelta=bn[.,1:n]; @ matrix to derive contribution of deltas to BN  Deviations @  
ydevdelta=(bndelta*(alldy-growth')')'; @ contribution of delta terms to BN deviations@

RETP(bntrend,ydev,bndelta,ydevdelta,forclevels,
     ally,alldy,allz,growth);
ENDP;



proc(1)= power(b,i,g);
local jj, bi;
bi=eye(g);
jj=0;
do while jj<i;
bi=bi*b;
jj=jj+1;
endo;
retp(bi);
endp;

